/*
  This file is part of the eWBM LoRa library.

  Based on the TinyGSM library https://github.com/vshymanskyy/TinyGSM
  Copyright (c) 2016 Volodymyr Shymanskyy

  eWBM LoRa library is free software: you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  eWBM LoRa library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with MKR WAN library.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "Arduino.h"
#include "USARTClass.h"

#define LORA_RESET_PORT 8
USARTClass* g_pSerial = NULL;

#ifdef LORA_DEBUG
namespace {
	void SetDebug(USARTClass* pSerial)
	{
		g_pSerial = pSerial;
	}

	void DEBUG(const char* fmt, ...)
	{
		if (NULL == g_pSerial)
			return;
		char buf[512] = { 0, };
		memset(buf, 0, 512);
		va_list args;

		va_start(args, fmt);
		vsprintf((char*)buf, fmt, args);
		va_end(args);

		//Serial1.print(buf);
		g_pSerial->print(buf);
	}

	template<typename T>
	static void DBG(T last) {
		if (NULL == g_pSerial)
			return;
		g_pSerial->print(last);
	}

	template<typename T, typename... Args>
	static void DBG(T head, Args... tail) {
		if (NULL == g_pSerial)
			return;
		g_pSerial->print(head);
		g_pSerial->print(' ');
		DBG(tail...);
	}

	void DEBUGS(String& strData, const char* fmt, ...)
	{
		char buf[512] = { 0, };
		memset(buf, 0, 512);
		va_list args;
		va_start(args, fmt);
		vsprintf((char*)buf, fmt, args);
		va_end(args);

		String strDbg;
		strDbg += buf;
		strDbg += " [";
		strDbg += strData;
		strDbg += "]";
		strDbg += "\r\n";
		DBG(strDbg);
	}
}
#else
#define DBG(...)
#define DEBUG(...)
#define DEBUGS(...)
#define SetDebug
#endif



template <class T, unsigned N>
class SerialFifo
{
public:
	SerialFifo()
	{
		clear();
	}

	void clear()
	{
		_r = 0;
		_w = 0;
	}

	// writing thread/context API
	//-------------------------------------------------------------

	bool writeable(void)
	{
		return free() > 0;
	}

	int free(void)
	{
		int s = _r - _w;
		if (s <= 0)
			s += N;
		return s - 1;
	}

	bool put(const T& c)
	{
		int i = _w;
		int j = i;
		i = _inc(i);
		if (i == _r) // !writeable()
			return false;
		_b[j] = c;
		_w = i;
		return true;
	}

	int put(const T* p, int n, bool t = false)
	{
		int c = n;
		while (c)
		{
			int f;
			while ((f = free()) == 0) // wait for space
			{
				if (!t) return n - c; // no more space and not blocking
				/* nothing / just wait */;
			}
			// check free space
			if (c < f) f = c;
			int w = _w;
			int m = N - w;
			// check wrap
			if (f > m) f = m;
			memcpy(&_b[w], p, f);
			_w = _inc(w, f);
			c -= f;
			p += f;
		}
		return n - c;
	}

	// reading thread/context API
	// --------------------------------------------------------

	bool readable(void)
	{
		return (_r != _w);
	}

	size_t size(void)
	{
		int s = _w - _r;
		if (s < 0)
			s += N;
		return s;
	}

	bool get(T* p)
	{
		int r = _r;
		if (r == _w) // !readable()
			return false;
		*p = _b[r];
		_r = _inc(r);
		return true;
	}

	bool peek(T* p)
	{
		int r = _r;
		if (r == _w) // !readable()
			return false;
		*p = _b[r];
		return true;
	}

	int get(T* p, int n, bool t = false)
	{
		int c = n;
		while (c)
		{
			int f;
			for (;;) // wait for data
			{
				f = size();
				if (f)  break;        // free space
				if (!t) return n - c; // no space and not blocking
				/* nothing / just wait */;
			}
			// check available data
			if (c < f) f = c;
			int r = _r;
			int m = N - r;
			// check wrap
			if (f > m) f = m;
			memcpy(p, &_b[r], f);
			_r = _inc(r, f);
			c -= f;
			p += f;
		}
		return n - c;
	}

private:
	int _inc(int i, int n = 1)
	{
		return (i + n) % N;
	}

	T    _b[N];
	int  _w;
	int  _r;
};



#ifndef YIELD
#define YIELD() { delay(1); }
#endif

typedef const char* ConstStr;
#define GFP(x) x
#define GF(x)  x

#if !defined(LORA_RX_BUFFER)
#define LORA_RX_BUFFER 256
#endif

#define LORA_NL "\r\n"
static const char LORA_OK[] = "<OK>";
static const char LORA_ERROR[] = "<Failed";

typedef enum {
	ABP = 0,
	OTAA,
} _lora_mode;

typedef enum {
	APP_EUI = 0,
	APP_KEY,
	DEV_EUI,
	DEV_ADDR,
	NWKS_KEY,
	APPS_KEY,
	NWK_ID,
} _lora_property;

typedef enum {
	CLASS_A = 'A',
	CLASS_B,
	CLASS_C,
} _lora_class;

class LoRaModem : public Stream
{

public:
	LoRaModem(__attribute__((unused)) Stream& stream = (Stream&)Serial)
		: stream(stream), lastPollTime(millis()), pollInterval(300000)
	{
	}

public:
	typedef SerialFifo<uint8_t, LORA_RX_BUFFER> RxFifo;


protected:
	Stream&       stream;
	bool          network_joined;
	RxFifo        rx;
	RxFifo        tx;
	String        fw_version;
	unsigned long lastPollTime;
	unsigned long pollInterval;
	uint8_t		m_port;

public:
	void ResetHW() {
		pinMode(LORA_RESET_PORT, INPUT);
		pinMode(LORA_RESET_PORT, OUTPUT);
		digitalWrite(LORA_RESET_PORT, LOW);
		delay(1);
		pinMode(LORA_RESET_PORT, INPUT);
		delay(1000);
	}

	size_t write(uint8_t c) {
		return tx.put(c);
	};

	size_t write(const uint8_t *buffer, size_t size) {
		DEBUG("write:%s, SIZE=%d, (write)\r\n", (char*)buffer, (int)size);
		return tx.put(buffer, size);
	};

	template <typename T> inline size_t write(T val) { return write((uint8_t*)&val, sizeof(T)); };
	using Print::write;

	void beginPacket() {
		tx.clear();
	}

	bool setPort(uint8_t port) {
		m_port = port;
		return true;
	}

	int endPacket(bool IsBin) {
		uint8_t buffer[LORA_RX_BUFFER];
		memset(buffer, 0, LORA_RX_BUFFER);
		int size = tx.get(buffer, tx.size());
		DEBUG("endPackt:%s,%d\r\n", (char*)buffer, size);
		return modemSend(buffer, size, IsBin);
	}

	int SendPacket(int port, String str, bool IsBin)
	{
		int nRet;
		setPort(port);
		beginPacket();
		print(str);
		nRet = endPacket(IsBin);
		return nRet;
	}

#if 1
	int parsePacket() {
		return available();
	}

	virtual int available() {
		YIELD();
		if (!rx.size()) {
			maintain();
		}
		return rx.size(); // + buf_available;
	}

	virtual int read(uint8_t *buf, size_t size) {
		YIELD();
		maintain();
		size_t cnt = 0;
		while (cnt < size) {
			size_t chunk = Min(size - cnt, rx.size());
			if (chunk > 0) {
				rx.get(buf, chunk);
				buf += chunk;
				cnt += chunk;
				continue;
			}
			// TODO: Read directly into user buffer?
			maintain();
			/*
			if (buf_available > 0) {
			modemRead(rx.free());
			} else {
			break;
			}
			*/
		}
		return cnt;
	}

	virtual int read() {
		uint8_t c;
		if (read(&c, 1) == 1) {
			return c;
		}
		return -1;
	}

	virtual int peek() {
		uint8_t c;
		if (rx.peek(&c) == true) {
			return c;
		}
		return -1;
	}

	virtual uint8_t connected() {
		if (available()) {
			return true;
		}
		return network_joined;
	}

	virtual operator bool() { return connected(); }
#endif

	virtual void flush() { stream.flush(); }

public:
	bool autoBaud(unsigned long timeout = 10000L) {
		for (unsigned long start = millis(); millis() - start < timeout; ) {
			sendAT(GF(""));
			if (waitResponse(200) == 1) {
				delay(100);
				return true;
			}
			delay(100);
		}
		return false;
	}

	void maintain() {
		while (stream.available()) {
			waitResponse(100);
		}
	}

	void minPollInterval(unsigned long secs) {
		pollInterval = secs * 1000;
	}

	String GetRecv(int TimeOut = 1000) {
		String Recv;
		String data;
		if (waitResponse(TimeOut, data, "<OK>") == 1) {
			Recv = data;
		}
		return Recv;
	}

	String GetRecv(String str1, int TimeOut = 1000 ) {
		String Recv;
		String data;
		if (waitResponse(TimeOut, data, str1.c_str()) == 1) {
			Recv = data;
		}
		return Recv;
	}
	
//////////////////////////////////////////////////////
protected:
	int modemSend(const void* buff, size_t len, bool IsBin) {
		delay(40);
		DEBUG("%s:Buff=%s, len=%d \r\n", __func__, (char*)buff, len);
		String strSend, strPort(m_port), strBuf((const char *)buff);
		//String strSend("AT+SEND="), strPort(m_port), strBuf((const char *)buff);
		if (IsBin)
			strSend = "AT+SENDB=";
		else
			strSend = "AT+SEND=";
		strSend += strPort;
		strSend += ":";
		strSend += strBuf;
		strSend += "\r\n";

		DEBUGS(strSend, "%s:SendData", __func__);
		//stream.write((uint8_t*)strSend.c_str(), strSend.length());
		for (int i = 0; i<strSend.length(); i++)
		{
			//delay(40);
			delay(20);
			char c = strSend.charAt(i);
			stream.write(c);
			stream.flush();
		}

#if 0
		String data;
		if (waitResponse(10000, data, "<OK>") == 1) {
			DEBUG("%s:OK, [%s]\r\n", __func__, data.c_str());
		}
		else
		{
			DEBUG("%s:FAIL, [%s]\r\n", __func__, data.c_str());
			return -1;
		}
#endif

#if 1
		if (waitResponse(10000L, "<OK>") != 1) {
			DEBUG("%s:fail 1 \r\n", __func__);
			return -1;
		}
#endif
		return len;
	}

	/* Utilities */
	template<typename T>
	void streamWrite(T last) {
		String strlast = last;
		//DEBUG("streamWrite:%s\r\n", strlast.c_str());
		for (int i = 0; i<strlast.length(); i++)
		{
			stream.write(strlast.charAt(i));
			stream.flush();
			delay(20);
			//char c = strlast.charAt(i);
			//stream.write(c);
			//stream.flush();
			//Serial.write(c);
			//Serial.flush();
			//delay(40);
		}
		//stream.print(strlast);
	}

	template<typename T, typename... Args>
	void streamWrite(T head, Args... tail) {
		//stream.print(head);
		String str(head);
		streamWrite(str);
		streamWrite(tail...);
	}

	int streamRead() { return stream.read(); }

	bool streamSkipUntil(char c) { //TODO: timeout
		while (true) {
			while (!stream.available()) {}
			if (stream.read() == c)
				return true;
		}
		return false;
	}

	template<typename... Args>
	void sendAT(Args... cmd) {
		streamWrite("AT", cmd..., LORA_NL);
		stream.flush();
		YIELD();
	}

	// TODO: Optimize this!
	uint8_t waitResponse(uint32_t timeout, String& data,
		ConstStr r1 = GFP(LORA_OK), ConstStr r2 = GFP(LORA_ERROR),
		ConstStr r3 = NULL, ConstStr r4 = NULL, ConstStr r5 = NULL)
	{
		data.reserve(256);
		int index = -1;
		int length = 0;
		unsigned long startMillis = millis();
		do {
			YIELD();
			while (stream.available() > 0) {
				int a = streamRead();
				//DEBUG("%c", a);
				if (a < 0) continue;
				data += (char)a;
				if (r1 && data.endsWith(r1)) {
					index = 1;
					goto finish;
				}
				else if (r2 && data.endsWith(r2)) {
					index = 2;
					goto finish;
				}
				else if (r3 && data.endsWith(r3)) {
					index = 3;
					goto finish;
				}
				else if (r4 && data.endsWith(r4)) {
					index = 4;
					goto finish;
				}
				else if (r5 && data.endsWith(r5)) {
					index = 5;
					goto finish;
				}
#if 0
				else if (data.endsWith("RECV")) {
					data = "";
					stream.readStringUntil(',').toInt();
					length = stream.readStringUntil('\r').toInt();
					streamSkipUntil('\n');
					streamSkipUntil('\n');
					for (int i = 0; i < length;) {
						if (stream.available()) {
							DEBUG("waitResponse rx.put \r\n");
							rx.put(stream.read());
							i++;
						}
					}
				}
#endif
			}
		} while (millis() - startMillis < timeout);
	finish:
		if (!index) {
			data.trim();
			if (data.length()) {
				DBG("### Unhandled:", data);
			}
			data = "";
		}
		return index;
	}

	uint8_t waitResponse(uint32_t timeout,
		ConstStr r1 = GFP(LORA_OK), ConstStr r2 = GFP(LORA_ERROR),
		ConstStr r3 = NULL, ConstStr r4 = NULL, ConstStr r5 = NULL)
	{
		String data;
		return waitResponse(timeout, data, r1, r2, r3, r4, r5);
	}

	uint8_t waitResponse(ConstStr r1 = GFP(LORA_OK), ConstStr r2 = GFP(LORA_ERROR),
		ConstStr r3 = NULL, ConstStr r4 = NULL, ConstStr r5 = NULL)
	{
		return waitResponse(1000, r1, r2, r3, r4, r5);
	}

};


class eWBMLoRaModem : public LoRaModem
{
private:
	bool Set_AT_NJM(_lora_mode mode) {
		sendAT(GF("+NJM="), mode);
		if (waitResponse() != 1) {
			return false;
		}
		return true;
	}

	size_t Get_AT_NJS() {
		sendAT(GF("+NJS"));
		if (waitResponse(2000L) != 1) {
			DEBUG("%s, fail\r\n", __func__);
			return 0;
		}
		DEBUG("%s, ok\r\n", __func__);
		return 1;
	}

	bool Set_AT_JOIN() {
		sendAT(GF("+JOIN"));
		if (waitResponse(60000L, "JOINED") != 1) {
			return false;
		}
		return true;
	}

	bool Set_AT(_lora_property prop, const char* value) {
		switch (prop) {
		case APP_EUI:
			sendAT(GF("+APPEUI="), value);
			break;
		case APP_KEY:
			sendAT(GF("+APPKEY="), value);
			break;
		case DEV_EUI:
			sendAT(GF("+DEUI="), value);
			break;
		case DEV_ADDR:
			sendAT(GF("+DADDR="), value);
			break;
		case NWKS_KEY:
			sendAT(GF("+NWKSKEY="), value);
			break;
		case NWK_ID:
			sendAT(GF("+IDNWK="), value);
			break;
		case APPS_KEY:
			sendAT(GF("+APPSKEY="), value);
			break;
		default:
			return false;
		}
		if (waitResponse() != 1) {
			return false;
		}
		return true;
	}

public:
	virtual int joinOTAA(const char *appEui, const char *appKey, const char *devEui = NULL) {
		YIELD();
		rx.clear();
		Set_AT_NJM(OTAA);
		Set_AT(APP_EUI, appEui);
		Set_AT(APP_KEY, appKey);
		if (devEui != NULL) {
			Set_AT(DEV_EUI, devEui);
		}
		network_joined = Set_AT_JOIN();
		delay(1000);
		return network_joined;
	}

	virtual int joinOTAA(String appEui, String appKey) {
		return joinOTAA(appEui.c_str(), appKey.c_str(), NULL);
	}

	virtual int joinOTAA(String appEui, String appKey, String devEui) {
		return joinOTAA(appEui.c_str(), appKey.c_str(), devEui.c_str());
	}
	virtual int joinABP(/*const char* nwkId, */const char * devAddr, const char * nwkSKey, const char * appSKey) {
		YIELD();
		rx.clear();
		Set_AT_NJM(ABP);
		Set_AT(DEV_ADDR, devAddr);
		Set_AT(NWKS_KEY, nwkSKey);
		Set_AT(APPS_KEY, appSKey);
		network_joined = Set_AT_JOIN();

		return (Get_AT_NJS() == 1);
	}

	virtual int joinABP(/*String nwkId, */String devAddr, String nwkSKey, String appSKey) {
		return joinABP(/*nwkId.c_str(), */devAddr.c_str(), nwkSKey.c_str(), appSKey.c_str());
	}

public:
	bool begin() {
		if (init()) {
			return true;
			//return configureBand(band);
		}
		return false;
	}

	bool init() {
		if (!autoBaud()) {
			return false;
		}
		return true;
	}

	void AT(String strATCMD) {
		sendAT(GF(strATCMD.c_str()));
	}

	bool IsRecved(String& strRecvData, int TimeOut = 10000)
	{
		String FIND_END="\r\n";
		String dataRaw;
		dataRaw = GetRecv(FIND_END, TimeOut);
		while (dataRaw.length() > 0)
		{
			//DBG("RAW: [" + dataRaw + "]");
			int nIndex = dataRaw.lastIndexOf("RECVB");
			if (nIndex != -1)
			{
				String RecvB;
				if (dataRaw[dataRaw.length() - 1] == '\n')
					dataRaw.remove(dataRaw.length() - 1);
				if (dataRaw[dataRaw.length() - 1] == '\r')
					dataRaw.remove(dataRaw.length() - 1);
				RecvB = dataRaw.substring(nIndex + 6, dataRaw.length());
				//DBG("Received: [" + RecvB + "]");
				strRecvData = RecvB;
				return true;
			}

			dataRaw = GetRecv(FIND_END, TimeOut);
		}

		return false;
	}

public:
	bool Set_AT(String ATcmd) {
		sendAT(ATcmd.c_str());
		if (waitResponse() != 1) {
			return false;
		}
		return true;
	}

	String Set_AT_Manual(String ATcmd) {
		sendAT(ATcmd.c_str());
		return GetRecv(1000);
	}

	bool Set_AT_RESET() {
		if (!autoBaud()) {
			return false;
		}
		sendAT(GF("+RESET"));
		if (waitResponse(10000L, "<OK>") != 1) {
			return false;
		}
		delay(2000);
		return init();
	}

	bool Set_AT_NWK(bool publicNetwork) {
		sendAT(GF("+NWK="), publicNetwork);
		if (waitResponse() != 1) {
			return false;
		}
		return true;
	}
	
	String Get_AT_VER() {
		sendAT(GF("+VER"));
		String data;
		if (waitResponse(2000, data, "<OK>") == 1) {
			data.remove(0, 4);
			data.remove(data.length() - 8, 8);
			fw_version = data;
		}
		return fw_version;
	}

	bool Set_AT_CLASS(_lora_class _class) {
		sendAT(GF("+CLASS="), (char)_class);
		if (waitResponse() != 1) {
			return false;
		}
		return true;
	}

	String Get_AT_DEVEUI() {
		String eui;
		sendAT(GF("+DEVEUI"));
		String data;
		if (waitResponse(2000, data, "<OK>") == 1) {
			data.remove(0, 4);
			data.remove(data.length() - 8, 8);
			eui = data;
		}
		else
			eui = "";
		return eui;
	}

	bool Set_AT_DATARATE(uint8_t dr) {
		sendAT(GF("+DATARATE="), dr);
		if (waitResponse() != 1) {
			return false;
		}
		return true;
	}

	bool Set_AT_ADR(bool adr) {
		sendAT(GF("+ADR="), adr);
		if (waitResponse() != 1) {
			return false;
		}
		return true;
	}

	bool Set_AT_CFM(bool confirm) {
		sendAT(GF("+CFM="), confirm);
		if (waitResponse() != 1) {
			return false;
		}
		return true;
	}


};